import typing as tp
from concurrent.futures import Future

from satella.coding.decorators import wraps

try:
    from opentracing import Span
except ImportError:
    class Span:
        pass

try:
    from cassandra.cluster import ResponseFuture
except ImportError:
    class ResponseFuture:
        pass


def trace_function(tracer, name: str, tags: tp.Optional[dict] = None):
    """
    Return a decorator that will trace the execution of a given function
    using tracer.start_active_span

    :param tracer: tracer to use
    :param name: Name of the trace
    :param tags: optional tags to use
    """
    def outer(fun):
        @wraps(fun)
        def inner(*args, **kwargs):
            with tracer.start_active_span(name, tags=tags):
                return fun(*args, **kwargs)
        return inner
    return outer


def trace_future(future: tp.Union[ResponseFuture, Future], span: Span):
    """
    Install a handler that will close a span upon a future completing,
    attaching the exception contents if the future ends with an exception.

    :param future: can be either a normal Future or a Cassandra's ResponseFuture
    :param span: span to close on future's completion
    """
    if isinstance(future, ResponseFuture):
        def close_exception(exc):
            # noinspection PyProtectedMember
            Span._on_error(span, type(exc), exc, '<unavailable>')
            span.finish()

        future.add_callback(span.finish)
        future.add_errback(close_exception)
    else:
        def close_future(fut):
            exc = fut.exception()
            if exc is not None:
                # noinspection PyProtectedMember
                Span._on_error(span, type(exc), exc, '<unavailable>')
            span.finish()
        future.add_done_callback(close_future)



import typing as tp
import inspect
from inspect import Parameter


def extract_type_signature_from(fun: tp.Callable) -> tp.Tuple[type, ...]:
    sign = []
    params = inspect.signature(fun).parameters
    for parameter in params.values():
        if parameter.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
            if parameter.annotation == Parameter.empty:
                sign.append(None)
            else:
                sign.append(parameter.annotation)
    return tuple(sign)


class overload:
    """
    A class used for method overloading.

    Note that methods can be only overloaded by their positional, or positional-and-keyword
    arguments. Overload distinguishment will be done at the level of positional arguments only.

    Note that typing checks will be done via isinstance().

    Use like this:

    >>> @overload
    >>> def what_type(x: str):
    >>>     print('String')
    >>> @what_type.overload
    >>> def what_type(x: int):
    >>>     print('Int')
    """

    def __init__(self, fun: tp.Callable):
        self.type_signatures_to_functions = {
            extract_type_signature_from(fun): fun
        }  # type: tp.Dict[tp.Tuple[type, ...], tp.Callable]
        if hasattr(fun, '__doc__'):
            self.__doc__ = fun.__doc__

    def overload(self, fun):
        """
        :raises ValueError: this signature already has an overload
        """
        sign = extract_type_signature_from(fun)
        if sign in self.type_signatures_to_functions:
            f = self.type_signatures_to_functions[sign]
            raise ValueError('Method of this signature is already overloaded with %s' % (f, ))
        self.type_signatures_to_functions[sign] = fun
        return self

    def __call__(self, *args, **kwargs):
        """
        Call one of the overloaded functions.

        :raises TypeError: no type signature matched
        """
        for sign, fun in self.type_signatures_to_functions.items():
            try:
                for type_, arg in zip(sign, args):
                    if type_ is None:
                        continue
                    if not isinstance(arg, type_):
                        raise ValueError()

                return fun(*args, **kwargs)
            except ValueError:
                pass
        raise TypeError('No matching functions found')



